==> oobium_datastore_gen-1.0.0.tar.gz: lib/src/generators/models_feature_collection.dart <==
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelCollectionGenerator {
  final Model m;
  ModelCollectionGenerator(this.m);
  ModelCollectionGenerator.link(this.field) : m = field.linkedModel!;

  Schema get schema => m.schema;
  Model get p => field.model;
  late final ModelField field;
  late final ModelField linkedField = field.linkedField!;

  String compile() =>
'''
class ${m.namePlural} extends ModelCollection<${m.namePlural}, ${m.name}, ${m.idType}, ${m.refName}> {
  final _ModelFeature _feature;
  ${m.namePlural}._(this._feature) : super(
    _feature,
    () => ${m.namePlural}._(_feature),
    (s) => ${m.refName}._(_feature, s.id)
  );
  $accessors
  $fetch
  $find
  $put
}
${m.fields.where((f) => f.isHasMany).map((f) => ModelCollectionGenerator.link(f).compileWithParent()).join('\n')}
''';

  String compileWithParent() =>
'''
class ${p.name}${m.namePlural} extends ModelCollection<${p.name}${m.namePlural}, ${m.name}, ${m.idType}, ${m.refName}> {
  final _ModelFeature _feature;
  final ${p.refName} _${p.varName};
  ${p.name}${m.namePlural}._(this._feature, this._${p.varName}) : super(
    _feature,
    () => ${p.name}${m.namePlural}._(_feature, _${p.varName}),
    (s) => ${m.refName}._(_feature, s.id),
    (s) => s.${linkedField.name} == _${p.varName}.id
  );
  $accessors
  $fetchWithParent
  $findWithParent
  $putWithParent
}
''';

  String get accessors =>
'''
  ${m.refName}? operator [](${m.idType} id) => getOrNull(id);
  ${m.refName} get(${m.idType} id) => getOrNull(id) ?? (throw Exception('${m.name}(\$id) does not exist'));
  ${m.refName}? getOrNull(${m.idType} id) {
    final ref = ${m.refName}._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }
''';

  String get getOrPutParams => m.dataFields.isEmpty ? '{}' : '{${m.dataFields.map((f) => f.putParameter).join(',')}}';

  String get putParams => '{${m.isSingleton ? '' : '${m.putIdParameter},'}${m.dataFields.map((f) => f.putParameter).join(',')}}';

  String get putWithParentParams {
    final fields = m.realFields.where((f) => f != linkedField);
    return fields.isEmpty ? '' : '{${fields.map((f) => f.putParameter).join(',')}}';
  }

  String get putConstructor => [
    if(m.isExternal) 'id'
    else if(m.isInternal) 'id ?? ObjectId()',
    ...m.dataFields.map((f) => f.name)
  ].join(',');

  String get putWithParentConstructor => [
    if(m.isExternal) 'id'
    else if(m.isInternal) 'id ?? ObjectId()',
  ...m.dataFields.map((f) => (f == linkedField)
      ? '_${f.name}.id'
      : f.name)
  ].join(',');

  String get fetch => (m.canFetch && schema.has<RestFeature>()) ?
'''

  void expire({$expireParams}) => _feature<_RestFeature>().${m.adapterVarName}.expire($expireArgs);
  
  Future<void> fetch({$fetchParams}) => _feature<_RestFeature>().${m.adapterVarName}.fetch($fetchArgs);
''' : '';

  String get expireParams => [
    ...m.realFields.map((f) => [
      if(f.isNullable) 'bool? ${f.name}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.name}',
    ]).expand((e) => e),
  ].join(',');

  String get expireArgs => [
    '${m.queryName}._($fetchQueryArgs)',
  ].join(',');

  String get fetchParams => [
    ...m.realFields.map((f) => [
      if(f.isNullable) 'bool? ${f.name}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.name}',
    ]).expand((e) => e),
    'bool force=false'
  ].join(',');

  String get fetchArgs => [
    '${m.queryName}._($fetchQueryArgs)',
    'force'
  ].join(',');

  String get fetchQueryArgs => m.realFields.map((f) => [
    if(f.isNullable) '${f.name}IsNull',
    f.name,
  ]).expand((e) => e).join(',');

  String get fetchWithParent => (m.canFetch && schema.has<RestFeature>()) ?
'''
  Future<void> fetch({$fetchWithParentParams}) => _feature<_RestFeature>().${m.adapterVarName}.fetch($fetchWithParentArgs);
''' : '';

  String get fetchWithParentParams => [
    ...m.realFields.where((f) => f != linkedField).map((f) => [
      if(f.isNullable) 'bool? ${f.name}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.name}',
    ]).expand((e) => e),
    'bool force=false'
  ].join(',');

  String get fetchWithParentArgs => [
    '${m.queryName}._($fetchWithParentQueryArgs)',
    'force'
  ].join(',');

  String get fetchWithParentQueryArgs => m.realFields.map((f) => [
    if(f != linkedField && f.isNullable) '${f.name}IsNull',
    if(f != linkedField) f.name,
    if(f == linkedField) '_${f.name}.id',
  ]).expand((e) => e).join(',');

  String get find =>
'''
  Iterable<${m.refName}> find({$findParams}) =>
    _feature.getAll(${m.queryName}._($findQueryArgs).test).map((s) => ${m.refName}._(_feature, s.id));
''';

  String get findParams => m.realFields.map((f) => [
    if(f.isNullable) 'bool? ${f.name}IsNull',
    '${f.typeModel?.idType ?? f.type}? ${f.name}',
  ]).expand((e) => e).join(',');

  String get findQueryArgs => m.realFields.map((f) => [
    if(f.isNullable) '${f.name}IsNull',
    f.name,
  ]).expand((e) => e).join(',');

  String get findWithParent =>
'''
  Iterable<${m.refName}> find({$findWithParentParams}) =>
    _feature.getAll(${m.queryName}._($findWithParentQueryArgs).test).map((s) => ${m.refName}._(_feature, s.id));
''';

  String get findWithParentParams => m.realFields.where((f) => f != linkedField).map((f) => [
    if(f.isNullable) 'bool? ${f.name}IsNull',
    '${f.typeModel?.idType ?? f.type}? ${f.name}',
  ]).expand((e) => e).join(',');

  String get findWithParentQueryArgs => m.realFields.map((f) => [
    if(f != linkedField && f.isNullable) '${f.name}IsNull',
    if(f != linkedField) f.name,
    if(f == linkedField) '_${f.name}.id',
  ]).expand((e) => e).join(',');

  String get put =>
'''
  ${m.refName} put($putParams) => load(${m.name}._($putConstructor));
''';

  String get putWithParent =>
'''
  ${m.refName} put($putWithParentParams) => load(${m.name}._($putWithParentConstructor));
''';
}

extension CollectionModelX on Model {
  String get putIdParameter => isExternal ? 'required $idType id' : '$idType? id';
}

extension CollectionModelFieldX on ModelField {

  String get putParameter => '${isRequired ? 'required $putType' : '$putType$nullable'} $name${isInitialized ? '=$initializer' : ''}';

  String get putType => typeModel?.idType ?? type;
}

==> oobium_datastore_gen-1.1.1.tar.gz: lib/src/generators/models_feature_collection.dart <==
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelCollectionGenerator {
  final Model m;
  ModelCollectionGenerator(this.m);
  ModelCollectionGenerator.link(this.field) : m = field.linkedModel!;

  Schema get schema => m.schema;
  Model get p => field.model;
  late final ModelField field;
  late final ModelField linkedField = field.linkedField!;

  String compile() =>
'''
class ${m.namePlural} extends ModelCollection<${m.namePlural}, ${m.snapName}, ${m.idType}, ${m.refName}> {
  final _ModelFeature _feature;
  ${m.namePlural}._(this._feature) : super(
    _feature,
    () => ${m.namePlural}._(_feature),
    (s) => ${m.refName}._(_feature, s.id)
  );
  $accessors
  $fetch
  $find
  $put
}
${m.fields.where((f) => f.isHasMany).map((f) => ModelCollectionGenerator.link(f).compileWithParent()).join('\n')}
''';

  String compileWithParent() =>
'''
class ${p.name}${m.namePlural} extends ModelCollection<${p.name}${m.namePlural}, ${m.snapName}, ${m.idType}, ${m.refName}> {
  final _ModelFeature _feature;
  final ${p.refName} _${p.varName};
  ${p.name}${m.namePlural}._(this._feature, this._${p.varName}) : super(
    _feature,
    () => ${p.name}${m.namePlural}._(_feature, _${p.varName}),
    (s) => ${m.refName}._(_feature, s.id),
    (s) => s.${linkedField.snapName} == _${p.varName}.id
  );
  $accessors
  $fetchWithParent
  $findWithParent
  $putWithParent
}
''';

  String get accessors =>
'''
  ${m.refName}? operator [](${m.idType} id) => getOrNull(id);
  ${m.refName} get(${m.idType} id) => getOrNull(id) ?? (throw Exception('${m.name}(\$id) does not exist'));
  ${m.refName}? getOrNull(${m.idType} id) {
    final ref = ${m.refName}._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }
''';

  String get getOrPutParams => m.dataFields.isEmpty ? '{}' : '{${m.dataFields.map((f) => f.putParameter).join(',')}}';

  String get putParams => '{${m.isSingleton ? '' : '${m.putIdParameter},'}${m.dataFields.map((f) => f.putParameter).join(',')}}';

  String get putWithParentParams {
    final fields = m.realFields.where((f) => f != linkedField);
    return fields.isEmpty ? '' : '{${fields.map((f) => f.putParameter).join(',')}}';
  }

  String get putConstructor => [
    if(m.isExternal) 'id'
    else if(m.isInternal) 'id ?? ObjectId()',
    ...m.dataFields.map((f) => f.snapName)
  ].join(',');

  String get putWithParentConstructor => [
    if(m.isExternal) 'id'
    else if(m.isInternal) 'id ?? ObjectId()',
  ...m.dataFields.map((f) => (f == linkedField)
      ? '_${f.name}.id'
      : f.snapName)
  ].join(',');

  String get fetch => (m.canFetch && schema.has<RestFeature>()) ?
'''

  void expire({$expireParams}) => _feature<_RestFeature>().${m.restAdapterVarName}.expire($expireArgs);
  
  Future<void> fetch({$fetchParams}) => _feature<_RestFeature>().${m.restAdapterVarName}.fetch($fetchArgs);
''' : '';

  String get expireParams => [
    ...m.realFields.map((f) => [
      if(f.isNullable) 'bool? ${f.snapName}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.snapName}',
    ]).expand((e) => e),
  ].join(',');

  String get expireArgs => [
    '${m.queryName}._($fetchQueryArgs)',
  ].join(',');

  String get fetchParams => [
    ...m.realFields.map((f) => [
      if(f.isNullable) 'bool? ${f.snapName}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.snapName}',
    ]).expand((e) => e),
    'bool force=false'
  ].join(',');

  String get fetchArgs => [
    '${m.queryName}._($fetchQueryArgs)',
    'force'
  ].join(',');

  String get fetchQueryArgs => m.realFields.map((f) => [
    if(f.isNullable) '${f.snapName}IsNull',
    f.snapName,
  ]).expand((e) => e).join(',');

  String get fetchWithParent => (m.canFetch && schema.has<RestFeature>()) ?
'''
  Future<void> fetch({$fetchWithParentParams}) => _feature<_RestFeature>().${m.restAdapterVarName}.fetch($fetchWithParentArgs);
''' : '';

  String get fetchWithParentParams => [
    ...m.realFields.where((f) => f != linkedField).map((f) => [
      if(f.isNullable) 'bool? ${f.snapName}IsNull',
      '${f.typeModel?.idType ?? f.type}? ${f.snapName}',
    ]).expand((e) => e),
    'bool force=false'
  ].join(',');

  String get fetchWithParentArgs => [
    '${m.queryName}._($fetchWithParentQueryArgs)',
    'force'
  ].join(',');

  String get fetchWithParentQueryArgs => m.realFields.map((f) => [
    if(f != linkedField && f.isNullable) '${f.snapName}IsNull',
    if(f != linkedField) f.snapName,
    if(f == linkedField) '_${f.name}.id',
  ]).expand((e) => e).join(',');

  String get find =>
'''
  Iterable<${m.refName}> find({$findParams}) =>
    _feature.getAll(${m.queryName}._($findQueryArgs).test).map((s) => ${m.refName}._(_feature, s.id));
''';

  String get findParams => m.realFields.map((f) => [
    if(f.isNullable) 'bool? ${f.snapName}IsNull',
    '${f.typeModel?.idType ?? f.type}? ${f.snapName}',
  ]).expand((e) => e).join(',');

  String get findQueryArgs => m.realFields.map((f) => [
    if(f.isNullable) '${f.snapName}IsNull',
    f.snapName,
  ]).expand((e) => e).join(',');

  String get findWithParent =>
'''
  Iterable<${m.refName}> find({$findWithParentParams}) =>
    _feature.getAll(${m.queryName}._($findWithParentQueryArgs).test).map((s) => ${m.refName}._(_feature, s.id));
''';

  String get findWithParentParams => m.realFields.where((f) => f != linkedField).map((f) => [
    if(f.isNullable) 'bool? ${f.snapName}IsNull',
    '${f.typeModel?.idType ?? f.type}? ${f.snapName}',
  ]).expand((e) => e).join(',');

  String get findWithParentQueryArgs => m.realFields.map((f) => [
    if(f != linkedField && f.isNullable) '${f.snapName}IsNull',
    if(f != linkedField) f.snapName,
    if(f == linkedField) '_${f.name}.id',
  ]).expand((e) => e).join(',');

  String get put =>
'''
  ${m.refName} put($putParams) => load(${m.snapName}._($putConstructor));
''';

  String get putWithParent =>
'''
  ${m.refName} put($putWithParentParams) => load(${m.snapName}._($putWithParentConstructor));
''';
}

extension CollectionModelX on Model {
  String get putIdParameter => isExternal ? 'required $idType id' : '$idType? id';
}

extension CollectionModelFieldX on ModelField {

  String get putParameter => '${isRequired ? 'required $putType' : '$putType$nullable'} $snapName${isInitialized ? '=$initializer' : ''}';

  String get putType => typeModel?.idType ?? type;
}
